%%%----------------------------------------------------------------------
%%% File    : mlex_freeradius_conf.erl
%%% Author  :  <svg@surnet.ru>
%%% Purpose : Free Radius test grammar
%%% Created : 29 Jul 2002 by  <svg@surnet.ru>
%%%----------------------------------------------------------------------
%%%
%%% $Id$
%%%
%%% $Log$
%%%

-module(mlex_freeradius_conf).
-author('svg@surnet.ru').

-import(mlex,[c/1, nc/1, ci/1, ci/2, cni/1, cni/2, str/1, btw/3]).
-import(mlex,['@'/1, '|'/1, '*'/1, '+'/1, '?'/1, '.'/0, sol/1, eol/1]).

-export([grammar/0]).

-define(skip, fun (_, _, _) -> [] end).

grammar() ->
  mlex:grammar(
    [{ws,             ws(),             ?skip},
     {shell_comment,  shell_comment(),  ?skip},
     {true_kw,        true_kw(),        fun yeec_token/3},
     {false_kw,       false_kw(),       fun yeec_token/3},
     {decl_end,       decl_end(),       fun yeec_token/3},
     {assign_op,      assign_op(),      fun yeec_token/3},
     {r_brace,        r_brace(),        fun yeec_token/3},
     {l_brace,        l_brace(),        fun yeec_token/3},
     {float_num,      float_num(),      fun yeec_token/3},
     {integer_num,    integer_num(),    fun yeec_token/3},
     {ipv4_address,   ipv4_address(),   fun yeec_token/3},
     {ipv4_netmask,   ipv4_netmask(),   fun yeec_token/3},
     {tag_string,     tag_string(),     fun yeec_token/3},
     {unquote_string, unquote_string(), fun yeec_token/3},
     {dquote_string,  dquote_string(),  fun yeec_token/3},
     {squote_string,  squote_string(),  fun yeec_token/3}
    ]).

%% Keywords

true_kw() ->
  '|'([str("true"), str("ok"), str("yes")]).

false_kw() ->
  '|'([str("false"), str("no")]).

%% Operators

assign_op() -> c($=).
r_brace()   -> c($}).
l_brace()   -> c(${).
decl_end() -> '+'(ci("\r\n")).

%% Whitespace -- ignored

ws() ->	ci(" \t\f").

shell_comment() -> '@'([c($#), '*'('.'())]).

tag_string() -> '@'([lalpha(), '*'('|'([alnum(), ci("-_")]))]).

%% Strings

unquote_string() ->
  '+'('|'([alnum(), quote_chr(), punct()])).

dquote_string() ->
  '@'([c($"), '+'('|'([nc($"), quote_chr()])), c($")]).

squote_string() ->
  '@'([c($'), '+'('|'([nc($'), quote_chr()])), c($')]).

quote_chr() -> '@'([c($\\), cni("\f\r\n")]).

%% IPv4 address

ipv4_address() ->
  '@'([btw(3, 3, '@'([btw(1, 3, digit()), c($.)])), btw(1, 3, digit())]).

ipv4_netmask() ->
  '@'([ipv4_address(), c($/), '|'([ipv4_address(), btw(1, 2, digit())])]).

%%Float
%%(+|-)?[0-9]+\.[0-9]+((E|e)(+|-)?[0-9]+)?
float_num() ->
  '@'([integer_num(), c($.),'+'(digit()),'?'('@'([ci("Ee"), integer_num()]))]).

integer_num() ->
  '@'(['?'(ci("+-")), '+'(digit())]).

alnum() ->
  '|'([alpha(), digit()]).

digit() ->
  ci($0, $9).

alpha() ->
  '|'([lalpha(), ualpha()]).

lalpha() ->
  '|'([ci($a, $z), ci(223, 255)]).

ualpha() ->
  '|'([ci($A, $Z), ci(192, 222)]).

%% skipped " ' ; \ # { } =
punct() ->
  '|'([c($!), ci($\$, $&), ci($(, $/), c($:), 
       c($<), ci($>, $@), c($[), ci($^, $`), c($|), c($~)]).

%%
%% Format funs
%%
yeec_token(Class=integer_num, Line, Str) ->
  {Class, Line, list_to_integer(Str)};
yeec_token(Class=float_num, Line, Str) ->
  {Class, Line, list_to_float(Str)};
yeec_token(Class=true_kw, Line, _) ->
  {Class, Line, true};
yeec_token(bool, Line, _) ->
  {bool, Line, false};
yeec_token(Class=tag_string, Line, Str) ->
  {Class, Line, unquote_chars(Str)};
yeec_token(Class=dquote_string, Line, Str) ->
  {string, Line, unquote_chars(string:strip(Str, both, $"))};
yeec_token(Class=squote_string, Line, Str) ->
  {string, Line, unquote_chars(string:strip(Str, both, $'))};
yeec_token(Class, Line, Data) ->
  {Class, Line, Data}.

unquote_chars(Str) ->
  unquote_chars(Str, []).

unquote_chars([], Ret) ->
  io_lib:format(lists:reverse(Ret), []);
unquote_chars([$~|Rest], Acc) ->
  unquote_chars(Rest, [$~,$~|Acc]);
unquote_chars([C|Rest], Acc) ->
  unquote_chars(Rest, [C|Acc]).
