%%%----------------------------------------------------------------------
%%% File    : mlex_str_buf.erl
%%% Author  :  <svg@surnet.ru>
%%% Purpose : Mlex string buffer
%%% Created :  2 Sep 2002 by  <svg@surnet.ru>
%%%----------------------------------------------------------------------
%%%
%%% $Id$
%%%
%%% $Log$
%%%

-module(mlex_str_buf).
-author('svg@surnet.ru').

-export([forward/1, lnum/1, nxtchr/1, curchr/1, pos/1, pos/2, retchr/1]).
-export([ str_to_buf/1]).

-record(buf, {str=[],  lnum=0, pos=0, len=0}).

str_to_buf(Str) when list(Str) ->
  #buf{str=Str,  lnum=0, pos=0, len=length(Str)}.

curchr(B) ->
  {C, _} = nxtchr(retchr(B)),
  C.

nxtchr(B=#buf{ lnum=0, pos=0}) ->
  {sof, B#buf{lnum=1}};
nxtchr(B=#buf{pos=P, len=Len}) when Len == P ->
  {eof, B};
nxtchr(B=#buf{str=S, lnum=L, pos=P}) ->
  Chr = lists:nth(P+1, S),
  NL = if Chr == $\n -> L+1;
	  true -> L
       end,
  {Chr, B#buf{lnum=NL, pos=P+1}}.

retchr(B=#buf{pos=0}) ->
  B;
retchr(B=#buf{str=S, lnum=L, pos=P, len=Len}) when P =< Len ->
  Chr = lists:nth(P, S),
  NL = if Chr == 10 -> L-1;
	  true -> L
       end,
  B#buf{str=S, lnum=NL, pos=P-1}.

pos(B=#buf{pos=P}) ->
  P.

pos(B=#buf{str=S, pos=P}, NP) when NP == P ->
  B;
pos(B=#buf{str=S, pos=P}, NP) when NP < P ->
  pos(retchr(B), NP).

forward(B=#buf{len=0}) ->
  B;
forward(B=#buf{str=S,  pos=P, len=Len}) ->
  B#buf{str=lists:nthtail(P, S), pos=0, len=Len-P}.

lnum(B=#buf{lnum=L}) ->
  L.
